@@ -11,6 +11,7 @@
 #include "nsWrapperCache.h"
 #include "mozilla/dom/GamepadHapticActuatorBinding.h"
 #include "mozilla/dom/Gamepad.h"
+#include "mozilla/dom/GamepadHandle.h"
 
 namespace mozilla {
 namespace dom {
@@ -18,7 +19,7 @@ class Promise;
 
 class GamepadHapticActuator : public nsISupports, public nsWrapperCache {
  public:
-  GamepadHapticActuator(nsISupports* aParent, uint32_t aGamepadId,
+  GamepadHapticActuator(nsISupports* aParent, GamepadHandle aGamepadHandle,
                         uint32_t aIndex);
 
   NS_DECL_CYCLE_COLLECTING_ISUPPORTS
@@ -41,7 +42,7 @@ class GamepadHapticActuator : public nsISupports, public nsWrapperCache {
 
  protected:
   nsCOMPtr<nsISupports> mParent;
-  uint32_t mGamepadId;
+  GamepadHandle mGamepadHandle;
   GamepadHapticActuatorType mType;
   uint32_t mIndex;
 };
