@@ -9,9 +9,7 @@
 #include "OMX_Audio.h"
 #include "OMX_Component.h"
 #include "OMX_Types.h"
-
 #include "OmxPlatformLayer.h"
-
 #include "mozilla/IntegerPrintfMacros.h"
 
 #ifdef LOG
@@ -35,6 +33,8 @@
 
 namespace mozilla {
 
+using namespace gfx;
+
 static const char* StateTypeToStr(OMX_STATETYPE aType) {
   MOZ_ASSERT(aType == OMX_StateLoaded || aType == OMX_StateIdle ||
              aType == OMX_StateExecuting || aType == OMX_StatePause ||
@@ -61,7 +61,7 @@ static const char* StateTypeToStr(OMX_STATETYPE aType) {
 // A helper class to retrieve AudioData or VideoData.
 class MediaDataHelper {
  protected:
-  virtual ~MediaDataHelper() {}
+  virtual ~MediaDataHelper() = default;
 
  public:
   NS_INLINE_DECL_THREADSAFE_REFCOUNTING(MediaDataHelper)
@@ -92,11 +92,9 @@ class MediaDataHelper {
 };
 
 OmxDataDecoder::OmxDataDecoder(const TrackInfo& aTrackInfo,
-                               TaskQueue* aTaskQueue,
                                layers::ImageContainer* aImageContainer)
     : mOmxTaskQueue(
           CreateMediaDecodeTaskQueue("OmxDataDecoder::mOmxTaskQueue")),
-      mTaskQueue(aTaskQueue),
       mImageContainer(aImageContainer),
       mWatchManager(this, mOmxTaskQueue),
       mOmxState(OMX_STATETYPE::OMX_StateInvalid, "OmxDataDecoder::mOmxState"),
@@ -133,6 +131,7 @@ void OmxDataDecoder::EndOfStream() {
 RefPtr<MediaDataDecoder::InitPromise> OmxDataDecoder::Init() {
   LOG("");
 
+  mThread = GetCurrentSerialEventTarget();
   RefPtr<OmxDataDecoder> self = this;
   return InvokeAsync(mOmxTaskQueue, __func__, [self, this]() {
     InitializationTask();
@@ -156,6 +155,7 @@ RefPtr<MediaDataDecoder::InitPromise> OmxDataDecoder::Init() {
 RefPtr<MediaDataDecoder::DecodePromise> OmxDataDecoder::Decode(
     MediaRawData* aSample) {
   LOG("sample %p", aSample);
+  MOZ_ASSERT(mThread->IsOnCurrentThread());
   MOZ_ASSERT(mInitPromise.IsEmpty());
 
   RefPtr<OmxDataDecoder> self = this;
@@ -174,6 +174,7 @@ RefPtr<MediaDataDecoder::DecodePromise> OmxDataDecoder::Decode(
 
 RefPtr<MediaDataDecoder::FlushPromise> OmxDataDecoder::Flush() {
   LOG("");
+  MOZ_ASSERT(mThread->IsOnCurrentThread());
 
   mFlushing = true;
 
@@ -182,6 +183,7 @@ RefPtr<MediaDataDecoder::FlushPromise> OmxDataDecoder::Flush() {
 
 RefPtr<MediaDataDecoder::DecodePromise> OmxDataDecoder::Drain() {
   LOG("");
+  MOZ_ASSERT(mThread->IsOnCurrentThread());
 
   RefPtr<OmxDataDecoder> self = this;
   return InvokeAsync(mOmxTaskQueue, __func__, [self]() {
@@ -193,6 +195,8 @@ RefPtr<MediaDataDecoder::DecodePromise> OmxDataDecoder::Drain() {
 
 RefPtr<ShutdownPromise> OmxDataDecoder::Shutdown() {
   LOG("");
+  // mThread may not be set if Init hasn't been called first.
+  MOZ_ASSERT(!mThread || mThread->IsOnCurrentThread());
 
   mShuttingDown = true;
 
@@ -267,7 +271,7 @@ RefPtr<ShutdownPromise> OmxDataDecoder::DoAsyncShutdown() {
             return ShutdownPromise::CreateAndReject(false, __func__);
           })
       ->Then(
-          mTaskQueue, __func__,
+          mThread, __func__,
           [self]() {
             self->mOmxTaskQueue->BeginShutdown();
             self->mOmxTaskQueue->AwaitShutdownAndIdle();
@@ -919,24 +923,28 @@ already_AddRefed<VideoData> MediaDataHelper::CreateYUV420VideoData(
   b.mPlanes[0].mWidth = width;
   b.mPlanes[0].mHeight = height;
   b.mPlanes[0].mStride = stride;
-  b.mPlanes[0].mOffset = 0;
   b.mPlanes[0].mSkip = 0;
 
   b.mPlanes[1].mData = yuv420p_u;
   b.mPlanes[1].mWidth = (width + 1) / 2;
   b.mPlanes[1].mHeight = (height + 1) / 2;
   b.mPlanes[1].mStride = (stride + 1) / 2;
-  b.mPlanes[1].mOffset = 0;
   b.mPlanes[1].mSkip = 0;
 
   b.mPlanes[2].mData = yuv420p_v;
   b.mPlanes[2].mWidth = (width + 1) / 2;
   b.mPlanes[2].mHeight = (height + 1) / 2;
   b.mPlanes[2].mStride = (stride + 1) / 2;
-  b.mPlanes[2].mOffset = 0;
   b.mPlanes[2].mSkip = 0;
 
   VideoInfo info(*mTrackInfo->GetAsVideoInfo());
+
+  auto maybeColorSpace = info.mColorSpace;
+  if (!maybeColorSpace) {
+    maybeColorSpace = Some(DefaultColorSpace({width, height}));
+  }
+  b.mYUVColorSpace = *maybeColorSpace;
+
   RefPtr<VideoData> data = VideoData::CreateAndCopyData(
       info, mImageContainer,
       0,                                     // Filled later by caller.
