@@ -23,6 +23,7 @@
 #include "prprf.h"  // leaving it for PR_vsnprintf()
 #include "mozilla/IntegerPrintfMacros.h"
 #include "mozilla/Sprintf.h"
+#include "VideoUtils.h"
 
 #include <algorithm>
 #include <numeric>
@@ -232,13 +233,12 @@ already_AddRefed<MediaTrackDemuxer> WebMDemuxer::GetTrackDemuxer(
   return e.forget();
 }
 
-nsresult WebMDemuxer::Reset(TrackInfo::TrackType aType) {
+void WebMDemuxer::Reset(TrackInfo::TrackType aType) {
   if (aType == TrackInfo::kVideoTrack) {
     mVideoPackets.Reset();
   } else {
     mAudioPackets.Reset();
   }
-  return NS_OK;
 }
 
 nsresult WebMDemuxer::ReadMetadata() {
@@ -364,8 +364,7 @@ nsresult WebMDemuxer::ReadMetadata() {
       if (mInfo.mVideo.mCrypto.IsEncrypted()) {
         MOZ_ASSERT(mInfo.mVideo.mCrypto.mCryptoScheme == CryptoScheme::Cenc,
                    "WebM should only use cenc scheme");
-        mCrypto.AddInitData(NS_LITERAL_STRING("webm"),
-                            mInfo.mVideo.mCrypto.mKeyId);
+        mCrypto.AddInitData(u"webm"_ns, mInfo.mVideo.mCrypto.mKeyId);
       }
     } else if (type == NESTEGG_TRACK_AUDIO && !mHasAudio) {
       nestegg_audio_params params;
@@ -435,8 +434,7 @@ nsresult WebMDemuxer::ReadMetadata() {
       if (mInfo.mAudio.mCrypto.IsEncrypted()) {
         MOZ_ASSERT(mInfo.mAudio.mCrypto.mCryptoScheme == CryptoScheme::Cenc,
                    "WebM should only use cenc scheme");
-        mCrypto.AddInitData(NS_LITERAL_STRING("webm"),
-                            mInfo.mAudio.mCrypto.mKeyId);
+        mCrypto.AddInitData(u"webm"_ns, mInfo.mAudio.mCrypto.mKeyId);
       }
     }
   }
@@ -461,7 +459,7 @@ void WebMDemuxer::EnsureUpToDateIndex() {
       Resource(TrackInfo::kVideoTrack).GetResource());
   MediaByteRangeSet byteRanges;
   nsresult rv = resource->GetCachedRanges(byteRanges);
-  if (NS_FAILED(rv) || !byteRanges.Length()) {
+  if (NS_FAILED(rv) || byteRanges.IsEmpty()) {
     return;
   }
   mBufferedState->UpdateIndex(byteRanges, resource);
@@ -641,8 +639,8 @@ nsresult WebMDemuxer::GetNextPacket(TrackInfo::TrackType aType,
             packetEncryption == NESTEGG_PACKET_HAS_SIGNAL_BYTE_UNENCRYPTED ||
                 packetEncryption == NESTEGG_PACKET_HAS_SIGNAL_BYTE_FALSE,
             "Unencrypted packet expected");
-        auto sample = MakeSpan(data, length);
-        auto alphaSample = MakeSpan(alphaData, alphaLength);
+        auto sample = Span(data, length);
+        auto alphaSample = Span(alphaData, alphaLength);
 
         switch (mVideoCodec) {
           case NESTEGG_CODEC_VP8:
@@ -899,9 +897,7 @@ nsresult WebMDemuxer::SeekInternal(TrackInfo::TrackType aType,
   uint32_t trackToSeek = mHasVideo ? mVideoTrack : mAudioTrack;
   uint64_t target = aTarget.ToNanoseconds();
 
-  if (NS_FAILED(Reset(aType))) {
-    return NS_ERROR_FAILURE;
-  }
+  Reset(aType);
 
   if (mSeekPreroll) {
     uint64_t startTime = 0;
@@ -949,6 +945,33 @@ nsresult WebMDemuxer::SeekInternal(TrackInfo::TrackType aType,
   return NS_OK;
 }
 
+bool WebMDemuxer::IsBufferedIntervalValid(uint64_t start, uint64_t end) {
+  if (start > end) {
+    // Buffered ranges are clamped to the media's start time and duration. Any
+    // frames with timestamps outside that range are ignored, see bug 1697641
+    // for more info.
+    WEBM_DEBUG("Ignoring range %" PRIu64 "-%" PRIu64
+               ", due to invalid interval (start > end).",
+               start, end);
+    return false;
+  }
+
+  auto startTime = TimeUnit::FromNanoseconds(start);
+  auto endTime = TimeUnit::FromNanoseconds(end);
+
+  if (startTime.IsNegative() || endTime.IsNegative()) {
+    // We can get timestamps that are conceptually valid, but become
+    // negative due to uint64 -> int64 conversion from TimeUnit. We should
+    // not get negative timestamps, so guard against them.
+    WEBM_DEBUG(
+        "Invalid range %f-%f, likely result of uint64 -> int64 conversion.",
+        startTime.ToSeconds(), endTime.ToSeconds());
+    return false;
+  }
+
+  return true;
+}
+
 media::TimeIntervals WebMDemuxer::GetBuffered() {
   EnsureUpToDateIndex();
   AutoPinned<MediaResource> resource(
@@ -985,8 +1008,15 @@ media::TimeIntervals WebMDemuxer::GetBuffered() {
                    TimeUnit::FromNanoseconds(duration).ToSeconds());
         end = duration;
       }
+
+      if (!IsBufferedIntervalValid(start, end)) {
+        WEBM_DEBUG("Invalid interval, bailing");
+        break;
+      }
+
       auto startTime = TimeUnit::FromNanoseconds(start);
       auto endTime = TimeUnit::FromNanoseconds(end);
+
       WEBM_DEBUG("add range %f-%f", startTime.ToSeconds(), endTime.ToSeconds());
       buffered += media::TimeInterval(startTime, endTime);
     }
@@ -1102,15 +1132,19 @@ RefPtr<WebMTrackDemuxer::SamplesPromise> WebMTrackDemuxer::GetSamples(
     if (mNeedKeyframe && !sample->mKeyframe) {
       continue;
     }
+    if (!sample->HasValidTime()) {
+      return SamplesPromise::CreateAndReject(NS_ERROR_DOM_MEDIA_DEMUXER_ERR,
+                                             __func__);
+    }
     mNeedKeyframe = false;
-    samples->mSamples.AppendElement(sample);
+    samples->AppendSample(sample);
     aNumSamples--;
   }
 
-  if (samples->mSamples.IsEmpty()) {
+  if (samples->GetSamples().IsEmpty()) {
     return SamplesPromise::CreateAndReject(rv, __func__);
   } else {
-    UpdateSamples(samples->mSamples);
+    UpdateSamples(samples->GetSamples());
     return SamplesPromise::CreateAndResolve(samples, __func__);
   }
 }
@@ -1177,7 +1211,7 @@ void WebMTrackDemuxer::Reset() {
   mSamples.Reset();
   media::TimeIntervals buffered = GetBuffered();
   mNeedKeyframe = true;
-  if (buffered.Length()) {
+  if (!buffered.IsEmpty()) {
     WEBM_DEBUG("Seek to start point: %f", buffered.Start(0).ToSeconds());
     mParent->SeekInternal(mType, buffered.Start(0));
     SetNextKeyFrameTime();
@@ -1186,7 +1220,8 @@ void WebMTrackDemuxer::Reset() {
   }
 }
 
-void WebMTrackDemuxer::UpdateSamples(nsTArray<RefPtr<MediaRawData>>& aSamples) {
+void WebMTrackDemuxer::UpdateSamples(
+    const nsTArray<RefPtr<MediaRawData>>& aSamples) {
   for (const auto& sample : aSamples) {
     if (sample->mCrypto.IsEncrypted()) {
       UniquePtr<MediaRawDataWriter> writer(sample->CreateWriter());
@@ -1254,6 +1289,6 @@ int64_t WebMTrackDemuxer::GetEvictionOffset(const TimeUnit& aTime) {
 
   return offset;
 }
+}  // namespace mozilla
 
 #undef WEBM_DEBUG
-}  // namespace mozilla
