@@ -4,32 +4,26 @@
  * file, You can obtain one at http://mozilla.org/MPL/2.0/. */
 
 #include "nsIAppShellService.h"
-#include "nsIComponentManager.h"
-#include "nsIURL.h"
 #include "nsNetUtil.h"
-#include "nsIServiceManager.h"
 #include "nsIObserverService.h"
 #include "nsIObserver.h"
-#include "nsIXPConnect.h"
 #include "nsIXULRuntime.h"
 
 #include "nsIWindowMediator.h"
-#include "nsIWindowWatcher.h"
 #include "nsPIWindowWatcher.h"
-#include "nsIDOMWindow.h"
 #include "nsPIDOMWindow.h"
-#include "nsWebShellWindow.h"
+#include "AppWindow.h"
 
 #include "nsWidgetInitData.h"
 #include "nsWidgetsCID.h"
 #include "nsIWidget.h"
-#include "nsIRequestObserver.h"
 #include "nsIEmbeddingSiteWindow.h"
 
+#include "nsAppDirectoryServiceDefs.h"
 #include "nsAppShellService.h"
 #include "nsContentUtils.h"
+#include "nsDirectoryServiceUtils.h"
 #include "nsThreadUtils.h"
-#include "nsISupportsPrimitives.h"
 #include "nsILoadContext.h"
 #include "nsIWebNavigation.h"
 #include "nsIWindowlessBrowser.h"
@@ -38,18 +32,25 @@
 #include "mozilla/Preferences.h"
 #include "mozilla/Services.h"
 #include "mozilla/StartupTimeline.h"
+#include "mozilla/StaticPrefs_fission.h"
 #include "mozilla/intl/LocaleService.h"
+#include "mozilla/dom/BrowsingContext.h"
 
 #include "nsEmbedCID.h"
 #include "nsIWebBrowser.h"
 #include "nsIDocShell.h"
 #include "gfxPlatform.h"
 
+#include "nsWebBrowser.h"
+#include "nsDocShell.h"
+#include "nsDocShellLoadState.h"
+
 #ifdef MOZ_INSTRUMENT_EVENT_LOOP
-#include "EventTracer.h"
+#  include "EventTracer.h"
 #endif
 
 using namespace mozilla;
+using mozilla::dom::BrowsingContext;
 using mozilla::intl::LocaleService;
 
 // Default URL for the hidden window, can be overridden by a pref on Mac
@@ -78,24 +79,36 @@ nsAppShellService::~nsAppShellService() {}
  */
 NS_IMPL_ISUPPORTS(nsAppShellService, nsIAppShellService, nsIObserver)
 
-NS_IMETHODIMP
-nsAppShellService::CreateHiddenWindow() {
-  return CreateHiddenWindowHelper(false);
-}
-
 NS_IMETHODIMP
 nsAppShellService::SetScreenId(uint32_t aScreenId) {
   mScreenId = aScreenId;
   return NS_OK;
 }
 
-void nsAppShellService::EnsurePrivateHiddenWindow() {
-  if (!mHiddenPrivateWindow) {
-    CreateHiddenWindowHelper(true);
+void nsAppShellService::EnsureHiddenWindow() {
+  if (!mHiddenWindow) {
+    (void)CreateHiddenWindow();
   }
 }
 
-nsresult nsAppShellService::CreateHiddenWindowHelper(bool aIsPrivate) {
+NS_IMETHODIMP
+nsAppShellService::CreateHiddenWindow() {
+  if (!XRE_IsParentProcess()) {
+    return NS_ERROR_NOT_IMPLEMENTED;
+  }
+
+  if (mXPCOMShuttingDown) {
+    return NS_ERROR_FAILURE;
+  }
+
+  nsCOMPtr<nsIFile> profileDir;
+  NS_GetSpecialDirectory(NS_APP_USER_PROFILE_50_DIR,
+                         getter_AddRefs(profileDir));
+  if (!profileDir) {
+    // This is too early on startup to create the hidden window
+    return NS_ERROR_FAILURE;
+  }
+
   nsresult rv;
   int32_t initialHeight = 100, initialWidth = 100;
 
@@ -105,11 +118,7 @@ nsresult nsAppShellService::CreateHiddenWindowHelper(bool aIsPrivate) {
   rv = Preferences::GetCString("browser.hiddenWindowChromeURL", prefVal);
   const char* hiddenWindowURL =
       NS_SUCCEEDED(rv) ? prefVal.get() : DEFAULT_HIDDENWINDOW_URL;
-  if (aIsPrivate) {
-    hiddenWindowURL = DEFAULT_HIDDENWINDOW_URL;
-  } else {
-    mApplicationProvidedHiddenWindow = prefVal.get() ? true : false;
-  }
+  mApplicationProvidedHiddenWindow = prefVal.get() ? true : false;
 #else
   static const char hiddenWindowURL[] = DEFAULT_HIDDENWINDOW_URL;
   uint32_t chromeMask = nsIWebBrowserChrome::CHROME_ALL;
@@ -119,30 +128,19 @@ nsresult nsAppShellService::CreateHiddenWindowHelper(bool aIsPrivate) {
   rv = NS_NewURI(getter_AddRefs(url), hiddenWindowURL);
   NS_ENSURE_SUCCESS(rv, rv);
 
-  if (aIsPrivate) {
-    chromeMask |= nsIWebBrowserChrome::CHROME_PRIVATE_WINDOW;
-  }
-
-  RefPtr<nsWebShellWindow> newWindow;
-  rv =
-      JustCreateTopWindow(nullptr, url, chromeMask, initialWidth, initialHeight,
-                          true, nullptr, nullptr, getter_AddRefs(newWindow));
+  RefPtr<AppWindow> newWindow;
+  rv = JustCreateTopWindow(nullptr, url, chromeMask, initialWidth,
+                           initialHeight, true, getter_AddRefs(newWindow));
   NS_ENSURE_SUCCESS(rv, rv);
 
   nsCOMPtr<nsIDocShell> docShell;
   newWindow->GetDocShell(getter_AddRefs(docShell));
   if (docShell) {
-    docShell->SetIsActive(false);
-    if (aIsPrivate) {
-      docShell->SetAffectPrivateSessionLifetime(false);
-    }
+    Unused << docShell->GetBrowsingContext()->SetExplicitActive(
+        dom::ExplicitActiveStatus::Inactive);
   }
 
-  if (aIsPrivate) {
-    mHiddenPrivateWindow.swap(newWindow);
-  } else {
-    mHiddenWindow.swap(newWindow);
-  }
+  mHiddenWindow.swap(newWindow);
 
   return NS_OK;
 }
@@ -155,12 +153,6 @@ nsAppShellService::DestroyHiddenWindow() {
     mHiddenWindow = nullptr;
   }
 
-  if (mHiddenPrivateWindow) {
-    mHiddenPrivateWindow->Destroy();
-
-    mHiddenPrivateWindow = nullptr;
-  }
-
   return NS_OK;
 }
 
@@ -168,26 +160,24 @@ nsAppShellService::DestroyHiddenWindow() {
  * Create a new top level window and display the given URL within it...
  */
 NS_IMETHODIMP
-nsAppShellService::CreateTopLevelWindow(
-    nsIXULWindow* aParent, nsIURI* aUrl, uint32_t aChromeMask,
-    int32_t aInitialWidth, int32_t aInitialHeight, nsITabParent* aOpeningTab,
-    mozIDOMWindowProxy* aOpenerWindow, nsIXULWindow** aResult)
-
-{
+nsAppShellService::CreateTopLevelWindow(nsIAppWindow* aParent, nsIURI* aUrl,
+                                        uint32_t aChromeMask,
+                                        int32_t aInitialWidth,
+                                        int32_t aInitialHeight,
+                                        nsIAppWindow** aResult) {
   nsresult rv;
 
   StartupTimeline::RecordOnce(StartupTimeline::CREATE_TOP_LEVEL_WINDOW);
 
-  RefPtr<nsWebShellWindow> newWindow;
+  RefPtr<AppWindow> newWindow;
   rv = JustCreateTopWindow(aParent, aUrl, aChromeMask, aInitialWidth,
-                           aInitialHeight, false, aOpeningTab, aOpenerWindow,
-                           getter_AddRefs(newWindow));
+                           aInitialHeight, false, getter_AddRefs(newWindow));
   newWindow.forget(aResult);
 
   if (NS_SUCCEEDED(rv)) {
     // the addref resulting from this is the owning addref for this window
     RegisterTopLevelWindow(*aResult);
-    nsCOMPtr<nsIXULWindow> parent;
+    nsCOMPtr<nsIAppWindow> parent;
     if (aChromeMask & nsIWebBrowserChrome::CHROME_DEPENDENT) parent = aParent;
     (*aResult)->SetZLevel(CalculateWindowZLevel(parent, aChromeMask));
   }
@@ -196,23 +186,22 @@ nsAppShellService::CreateTopLevelWindow(
 }
 
 /*
- * This class provides a stub implementation of nsIWebBrowserChrome2, as needed
+ * This class provides a stub implementation of nsIWebBrowserChrome, as needed
  * by nsAppShellService::CreateWindowlessBrowser
  */
-class WebBrowserChrome2Stub : public nsIWebBrowserChrome2,
-                              public nsIEmbeddingSiteWindow,
-                              public nsIInterfaceRequestor,
-                              public nsSupportsWeakReference {
+class WebBrowserChrome2Stub final : public nsIWebBrowserChrome,
+                                    public nsIEmbeddingSiteWindow,
+                                    public nsIInterfaceRequestor,
+                                    public nsSupportsWeakReference {
  protected:
   nsCOMPtr<nsIWebBrowser> mBrowser;
   virtual ~WebBrowserChrome2Stub() {}
 
  public:
-  explicit WebBrowserChrome2Stub(nsIWebBrowser* aBrowser)
-      : mBrowser(aBrowser) {}
+  void SetBrowser(nsIWebBrowser* aBrowser) { mBrowser = aBrowser; }
+
   NS_DECL_ISUPPORTS
   NS_DECL_NSIWEBBROWSERCHROME
-  NS_DECL_NSIWEBBROWSERCHROME2
   NS_DECL_NSIINTERFACEREQUESTOR
   NS_DECL_NSIEMBEDDINGSITEWINDOW
 };
@@ -220,7 +209,6 @@ class WebBrowserChrome2Stub : public nsIWebBrowserChrome2,
 NS_INTERFACE_MAP_BEGIN(WebBrowserChrome2Stub)
   NS_INTERFACE_MAP_ENTRY_AMBIGUOUS(nsISupports, nsIWebBrowserChrome)
   NS_INTERFACE_MAP_ENTRY(nsIWebBrowserChrome)
-  NS_INTERFACE_MAP_ENTRY(nsIWebBrowserChrome2)
   NS_INTERFACE_MAP_ENTRY(nsIInterfaceRequestor)
   NS_INTERFACE_MAP_ENTRY(nsISupportsWeakReference)
   NS_INTERFACE_MAP_ENTRY(nsIEmbeddingSiteWindow)
@@ -229,24 +217,6 @@ NS_INTERFACE_MAP_END
 NS_IMPL_ADDREF(WebBrowserChrome2Stub)
 NS_IMPL_RELEASE(WebBrowserChrome2Stub)
 
-NS_IMETHODIMP
-WebBrowserChrome2Stub::SetStatus(uint32_t aStatusType,
-                                 const char16_t* aStatus) {
-  return NS_OK;
-}
-
-NS_IMETHODIMP
-WebBrowserChrome2Stub::GetWebBrowser(nsIWebBrowser** aWebBrowser) {
-  NS_NOTREACHED("WebBrowserChrome2Stub::GetWebBrowser is not supported");
-  return NS_ERROR_NOT_IMPLEMENTED;
-}
-
-NS_IMETHODIMP
-WebBrowserChrome2Stub::SetWebBrowser(nsIWebBrowser* aWebBrowser) {
-  NS_NOTREACHED("WebBrowserChrome2Stub::SetWebBrowser is not supported");
-  return NS_ERROR_NOT_IMPLEMENTED;
-}
-
 NS_IMETHODIMP
 WebBrowserChrome2Stub::GetChromeFlags(uint32_t* aChromeFlags) {
   *aChromeFlags = 0;
@@ -255,25 +225,15 @@ WebBrowserChrome2Stub::GetChromeFlags(uint32_t* aChromeFlags) {
 
 NS_IMETHODIMP
 WebBrowserChrome2Stub::SetChromeFlags(uint32_t aChromeFlags) {
-  NS_NOTREACHED("WebBrowserChrome2Stub::SetChromeFlags is not supported");
-  return NS_ERROR_NOT_IMPLEMENTED;
-}
-
-NS_IMETHODIMP
-WebBrowserChrome2Stub::DestroyBrowserWindow() {
-  NS_NOTREACHED("WebBrowserChrome2Stub::DestroyBrowserWindow is not supported");
-  return NS_ERROR_NOT_IMPLEMENTED;
-}
-
-NS_IMETHODIMP
-WebBrowserChrome2Stub::SizeBrowserTo(int32_t aCX, int32_t aCY) {
-  NS_NOTREACHED("WebBrowserChrome2Stub::SizeBrowserTo is not supported");
+  MOZ_ASSERT_UNREACHABLE(
+      "WebBrowserChrome2Stub::SetChromeFlags is "
+      "not supported");
   return NS_ERROR_NOT_IMPLEMENTED;
 }
 
 NS_IMETHODIMP
 WebBrowserChrome2Stub::ShowAsModal() {
-  NS_NOTREACHED("WebBrowserChrome2Stub::ShowAsModal is not supported");
+  MOZ_ASSERT_UNREACHABLE("WebBrowserChrome2Stub::ShowAsModal is not supported");
   return NS_ERROR_NOT_IMPLEMENTED;
 }
 
@@ -284,15 +244,7 @@ WebBrowserChrome2Stub::IsWindowModal(bool* aResult) {
 }
 
 NS_IMETHODIMP
-WebBrowserChrome2Stub::ExitModalEventLoop(nsresult aStatus) {
-  NS_NOTREACHED("WebBrowserChrome2Stub::ExitModalEventLoop is not supported");
-  return NS_ERROR_NOT_IMPLEMENTED;
-}
-
-NS_IMETHODIMP
-WebBrowserChrome2Stub::SetStatusWithContext(uint32_t aStatusType,
-                                            const nsAString& aStatusText,
-                                            nsISupports* aStatusContext) {
+WebBrowserChrome2Stub::SetLinkStatus(const nsAString& aStatusText) {
   return NS_OK;
 }
 
@@ -333,9 +285,6 @@ WebBrowserChrome2Stub::SetDimensions(uint32_t flags, int32_t x, int32_t y,
   return NS_OK;
 }
 
-NS_IMETHODIMP
-WebBrowserChrome2Stub::SetFocus() { return NS_ERROR_NOT_IMPLEMENTED; }
-
 NS_IMETHODIMP
 WebBrowserChrome2Stub::GetVisibility(bool* aVisibility) {
   return NS_ERROR_NOT_IMPLEMENTED;
@@ -369,11 +318,15 @@ class BrowserDestroyer final : public Runnable {
         mBrowser(aBrowser),
         mContainer(aContainer) {}
 
+  static nsresult Destroy(nsIWebBrowser* aBrowser) {
+    nsCOMPtr<nsIBaseWindow> window(do_QueryInterface(aBrowser));
+    return window->Destroy();
+  }
+
   NS_IMETHOD
   Run() override {
     // Explicitly destroy the browser, in case this isn't the last reference.
-    nsCOMPtr<nsIBaseWindow> window = do_QueryInterface(mBrowser);
-    return window->Destroy();
+    return Destroy(mBrowser);
   }
 
  protected:
@@ -398,27 +351,12 @@ class WindowlessBrowser final : public nsIWindowlessBrowser,
     mInterfaceRequestor = do_QueryInterface(aBrowser);
   }
   NS_DECL_ISUPPORTS
+  NS_DECL_NSIWINDOWLESSBROWSER
   NS_FORWARD_SAFE_NSIWEBNAVIGATION(mWebNavigation)
   NS_FORWARD_SAFE_NSIINTERFACEREQUESTOR(mInterfaceRequestor)
 
-  NS_IMETHOD
-  Close() override {
-    NS_ENSURE_TRUE(!mClosed, NS_ERROR_UNEXPECTED);
-    NS_ASSERTION(
-        nsContentUtils::IsSafeToRunScript(),
-        "WindowlessBrowser::Close called when not safe to run scripts");
-
-    mClosed = true;
-
-    mWebNavigation = nullptr;
-    mInterfaceRequestor = nullptr;
-
-    nsCOMPtr<nsIBaseWindow> window = do_QueryInterface(mBrowser);
-    return window->Destroy();
-  }
-
- protected:
-  virtual ~WindowlessBrowser() {
+ private:
+  ~WindowlessBrowser() {
     if (mClosed) {
       return;
     }
@@ -429,11 +367,10 @@ class WindowlessBrowser final : public nsIWindowlessBrowser,
     // when it's safe to run scripts. If this was triggered by GC, it may
     // not always be safe to run scripts, in which cases we need to delay
     // destruction until it is.
-    nsCOMPtr<nsIRunnable> runnable = new BrowserDestroyer(mBrowser, mContainer);
-    nsContentUtils::AddScriptRunner(runnable);
+    auto runnable = MakeRefPtr<BrowserDestroyer>(mBrowser, mContainer);
+    nsContentUtils::AddScriptRunner(runnable.forget());
   }
 
- private:
   nsCOMPtr<nsIWebBrowser> mBrowser;
   nsCOMPtr<nsIWebNavigation> mWebNavigation;
   nsCOMPtr<nsIInterfaceRequestor> mInterfaceRequestor;
@@ -447,30 +384,56 @@ NS_IMPL_ISUPPORTS(WindowlessBrowser, nsIWindowlessBrowser, nsIWebNavigation,
                   nsIInterfaceRequestor)
 
 NS_IMETHODIMP
-nsAppShellService::CreateWindowlessBrowser(bool aIsChrome,
+WindowlessBrowser::Close() {
+  NS_ENSURE_TRUE(!mClosed, NS_ERROR_UNEXPECTED);
+  NS_ASSERTION(nsContentUtils::IsSafeToRunScript(),
+               "WindowlessBrowser::Close called when not safe to run scripts");
+
+  mClosed = true;
+
+  mWebNavigation = nullptr;
+  mInterfaceRequestor = nullptr;
+  return BrowserDestroyer::Destroy(mBrowser);
+}
+
+NS_IMETHODIMP
+WindowlessBrowser::GetBrowsingContext(BrowsingContext** aBrowsingContext) {
+  nsCOMPtr<nsIDocShellTreeItem> docShellTreeItem = do_QueryInterface(mBrowser);
+  if (!docShellTreeItem) {
+    return NS_ERROR_NOT_INITIALIZED;
+  }
+  return docShellTreeItem->GetBrowsingContextXPCOM(aBrowsingContext);
+}
+
+NS_IMETHODIMP
+WindowlessBrowser::GetDocShell(nsIDocShell** aDocShell) {
+  nsCOMPtr<nsIDocShell> docShell = do_GetInterface(mInterfaceRequestor);
+  if (!docShell) {
+    return NS_ERROR_NOT_INITIALIZED;
+  }
+  docShell.forget(aDocShell);
+  return NS_OK;
+}
+
+NS_IMETHODIMP
+nsAppShellService::CreateWindowlessBrowser(bool aIsChrome, uint32_t aChromeMask,
                                            nsIWindowlessBrowser** aResult) {
-  /* First, we create an instance of nsWebBrowser. Instances of this class have
-   * an associated doc shell, which is what we're interested in.
-   */
-  nsCOMPtr<nsIWebBrowser> browser = do_CreateInstance(NS_WEBBROWSER_CONTRACTID);
-  if (!browser) {
-    NS_ERROR("Couldn't create instance of nsWebBrowser!");
-    return NS_ERROR_FAILURE;
+  if (aChromeMask) {
+    MOZ_DIAGNOSTIC_ASSERT(aIsChrome, "Got chrome flags for non-chrome browser");
+    if (aChromeMask & ~(nsIWebBrowserChrome::CHROME_REMOTE_WINDOW |
+                        nsIWebBrowserChrome::CHROME_FISSION_WINDOW |
+                        nsIWebBrowserChrome::CHROME_PRIVATE_WINDOW)) {
+      NS_ERROR("Received unexpected chrome flags");
+      return NS_ERROR_FAILURE;
+    }
   }
 
-  /* Next, we set the container window for our instance of nsWebBrowser. Since
+  /* First, we set the container window for our instance of nsWebBrowser. Since
    * we don't actually have a window, we instead set the container window to be
    * an instance of WebBrowserChrome2Stub, which provides a stub implementation
-   * of nsIWebBrowserChrome2.
+   * of nsIWebBrowserChrome.
    */
-  RefPtr<WebBrowserChrome2Stub> stub = new WebBrowserChrome2Stub(browser);
-  browser->SetContainerWindow(stub);
-
-  nsCOMPtr<nsIWebNavigation> navigation = do_QueryInterface(browser);
-
-  nsCOMPtr<nsIDocShellTreeItem> item = do_QueryInterface(navigation);
-  item->SetItemType(aIsChrome ? nsIDocShellTreeItem::typeChromeWrapper
-                              : nsIDocShellTreeItem::typeContentWrapper);
+  RefPtr<WebBrowserChrome2Stub> stub = new WebBrowserChrome2Stub();
 
   /* A windowless web browser doesn't have an associated OS level window. To
    * accomplish this, we initialize the window associated with our instance of
@@ -487,14 +450,41 @@ nsAppShellService::CreateWindowlessBrowser(bool aIsChrome,
     NS_ERROR("Couldn't create instance of stub widget");
     return NS_ERROR_FAILURE;
   }
+
   nsresult rv =
       widget->Create(nullptr, 0, LayoutDeviceIntRect(0, 0, 0, 0), nullptr);
   NS_ENSURE_SUCCESS(rv, rv);
-  nsCOMPtr<nsIBaseWindow> window = do_QueryInterface(navigation);
-  window->InitWindow(0, widget, 0, 0, 0, 0);
-  window->Create();
 
-  nsISupports* isstub = NS_ISUPPORTS_CAST(nsIWebBrowserChrome2*, stub);
+  // Create a BrowsingContext for our windowless browser.
+  RefPtr<BrowsingContext> browsingContext = BrowsingContext::CreateIndependent(
+      aIsChrome ? BrowsingContext::Type::Chrome
+                : BrowsingContext::Type::Content);
+
+  if (aChromeMask & nsIWebBrowserChrome::CHROME_REMOTE_WINDOW) {
+    browsingContext->SetRemoteTabs(true);
+  }
+  if (aChromeMask & nsIWebBrowserChrome::CHROME_FISSION_WINDOW) {
+    browsingContext->SetRemoteSubframes(true);
+  }
+  if (aChromeMask & nsIWebBrowserChrome::CHROME_PRIVATE_WINDOW) {
+    browsingContext->SetPrivateBrowsing(true);
+  }
+
+  /* Next, we create an instance of nsWebBrowser. Instances of this class have
+   * an associated doc shell, which is what we're interested in.
+   */
+  nsCOMPtr<nsIWebBrowser> browser = nsWebBrowser::Create(
+      stub, widget, browsingContext, nullptr /* initialWindowChild */);
+
+  if (NS_WARN_IF(!browser)) {
+    NS_ERROR("Couldn't create instance of nsWebBrowser!");
+    return NS_ERROR_FAILURE;
+  }
+
+  // Make sure the container window owns the the nsWebBrowser instance.
+  stub->SetBrowser(browser);
+
+  nsISupports* isstub = NS_ISUPPORTS_CAST(nsIWebBrowserChrome*, stub);
   RefPtr<nsIWindowlessBrowser> result = new WindowlessBrowser(browser, isstub);
   nsCOMPtr<nsIDocShell> docshell = do_GetInterface(result);
   docshell->SetInvisible(true);
@@ -503,15 +493,15 @@ nsAppShellService::CreateWindowlessBrowser(bool aIsChrome,
   return NS_OK;
 }
 
-uint32_t nsAppShellService::CalculateWindowZLevel(nsIXULWindow* aParent,
+uint32_t nsAppShellService::CalculateWindowZLevel(nsIAppWindow* aParent,
                                                   uint32_t aChromeMask) {
   uint32_t zLevel;
 
-  zLevel = nsIXULWindow::normalZ;
+  zLevel = nsIAppWindow::normalZ;
   if (aChromeMask & nsIWebBrowserChrome::CHROME_WINDOW_RAISED)
-    zLevel = nsIXULWindow::raisedZ;
+    zLevel = nsIAppWindow::raisedZ;
   else if (aChromeMask & nsIWebBrowserChrome::CHROME_WINDOW_LOWERED)
-    zLevel = nsIXULWindow::loweredZ;
+    zLevel = nsIAppWindow::loweredZ;
 
 #ifdef XP_MACOSX
   /* Platforms on which modal windows are always application-modal, not
@@ -546,7 +536,7 @@ static bool CheckForFullscreenWindow() {
   if (!wm) return false;
 
   nsCOMPtr<nsISimpleEnumerator> windowList;
-  wm->GetXULWindowEnumerator(nullptr, getter_AddRefs(windowList));
+  wm->GetAppWindowEnumerator(nullptr, getter_AddRefs(windowList));
   if (!windowList) return false;
 
   for (;;) {
@@ -573,17 +563,16 @@ static bool CheckForFullscreenWindow() {
  * Just do the window-making part of CreateTopLevelWindow
  */
 nsresult nsAppShellService::JustCreateTopWindow(
-    nsIXULWindow* aParent, nsIURI* aUrl, uint32_t aChromeMask,
+    nsIAppWindow* aParent, nsIURI* aUrl, uint32_t aChromeMask,
     int32_t aInitialWidth, int32_t aInitialHeight, bool aIsHiddenWindow,
-    nsITabParent* aOpeningTab, mozIDOMWindowProxy* aOpenerWindow,
-    nsWebShellWindow** aResult) {
+    AppWindow** aResult) {
   *aResult = nullptr;
   NS_ENSURE_STATE(!mXPCOMWillShutDown);
 
-  nsCOMPtr<nsIXULWindow> parent;
+  nsCOMPtr<nsIAppWindow> parent;
   if (aChromeMask & nsIWebBrowserChrome::CHROME_DEPENDENT) parent = aParent;
 
-  RefPtr<nsWebShellWindow> window = new nsWebShellWindow(aChromeMask);
+  RefPtr<AppWindow> window = new AppWindow(aChromeMask);
 
 #ifdef XP_WIN
   // If the parent is currently fullscreen, tell the child to ignore persisted
@@ -608,6 +597,40 @@ nsresult nsAppShellService::JustCreateTopWindow(
   if (aChromeMask & nsIWebBrowserChrome::CHROME_SUPPRESS_ANIMATION)
     widgetInitData.mIsAnimationSuppressed = true;
 
+  if (aChromeMask & nsIWebBrowserChrome::CHROME_ALWAYS_ON_TOP)
+    widgetInitData.mAlwaysOnTop = true;
+
+  if (aChromeMask & nsIWebBrowserChrome::CHROME_FISSION_WINDOW)
+    widgetInitData.mFissionWindow = true;
+
+#ifdef MOZ_WIDGET_GTK
+  // Linux/Gtk PIP window support. It's Chrome Toplevel window, always on top
+  // and without any bar.
+  uint32_t pipMask = nsIWebBrowserChrome::CHROME_ALWAYS_ON_TOP |
+                     nsIWebBrowserChrome::CHROME_OPENAS_CHROME;
+  uint32_t barMask = nsIWebBrowserChrome::CHROME_MENUBAR |
+                     nsIWebBrowserChrome::CHROME_TOOLBAR |
+                     nsIWebBrowserChrome::CHROME_LOCATIONBAR |
+                     nsIWebBrowserChrome::CHROME_STATUSBAR;
+  if (widgetInitData.mWindowType == eWindowType_toplevel &&
+      ((aChromeMask & pipMask) == pipMask) && !(aChromeMask & barMask)) {
+    widgetInitData.mPIPWindow = true;
+  }
+#elif defined(XP_WIN)
+  // Windows PIP window support. It's Chrome dialog window, always on top
+  // and without any bar.
+  uint32_t pipMask = nsIWebBrowserChrome::CHROME_ALWAYS_ON_TOP |
+                     nsIWebBrowserChrome::CHROME_OPENAS_CHROME;
+  uint32_t barMask = nsIWebBrowserChrome::CHROME_MENUBAR |
+                     nsIWebBrowserChrome::CHROME_TOOLBAR |
+                     nsIWebBrowserChrome::CHROME_LOCATIONBAR |
+                     nsIWebBrowserChrome::CHROME_STATUSBAR;
+  if (widgetInitData.mWindowType == eWindowType_dialog &&
+      ((aChromeMask & pipMask) == pipMask) && !(aChromeMask & barMask)) {
+    widgetInitData.mPIPWindow = true;
+  }
+#endif
+
 #ifdef XP_MACOSX
   // Mac OS X sheet support
   // Adding CHROME_OPENAS_CHROME to sheetMask makes modal windows opened from
@@ -618,7 +641,7 @@ nsresult nsAppShellService::JustCreateTopWindow(
   uint32_t sheetMask = nsIWebBrowserChrome::CHROME_OPENAS_DIALOG |
                        nsIWebBrowserChrome::CHROME_MODAL |
                        nsIWebBrowserChrome::CHROME_OPENAS_CHROME;
-  if (parent && (parent != mHiddenWindow && parent != mHiddenPrivateWindow) &&
+  if (parent && (parent != mHiddenWindow) &&
       ((aChromeMask & sheetMask) == sheetMask)) {
     widgetInitData.mWindowType = eWindowType_sheet;
   }
@@ -649,6 +672,7 @@ nsresult nsAppShellService::JustCreateTopWindow(
       widgetInitData.mBorderStyle = static_cast<enum nsBorderStyle>(
           widgetInitData.mBorderStyle | eBorderStyle_close);
     if (aChromeMask & nsIWebBrowserChrome::CHROME_WINDOW_RESIZE) {
+      widgetInitData.mResizable = true;
       widgetInitData.mBorderStyle = static_cast<enum nsBorderStyle>(
           widgetInitData.mBorderStyle | eBorderStyle_resizeh);
       // only resizable windows get the maximize button (but not dialogs)
@@ -679,25 +703,20 @@ nsresult nsAppShellService::JustCreateTopWindow(
 
   widgetInitData.mRTL = LocaleService::GetInstance()->IsAppLocaleRTL();
 
-  nsresult rv = window->Initialize(
-      parent, center ? aParent : nullptr, aUrl, aInitialWidth, aInitialHeight,
-      aIsHiddenWindow, aOpeningTab, aOpenerWindow, widgetInitData);
+  nsresult rv =
+      window->Initialize(parent, center ? aParent : nullptr, aInitialWidth,
+                         aInitialHeight, aIsHiddenWindow, widgetInitData);
 
   NS_ENSURE_SUCCESS(rv, rv);
 
   // Enforce the Private Browsing autoStart pref first.
   bool isPrivateBrowsingWindow =
       Preferences::GetBool("browser.privatebrowsing.autostart");
-  bool isUsingRemoteTabs = mozilla::BrowserTabsRemoteAutostart();
 
   if (aChromeMask & nsIWebBrowserChrome::CHROME_PRIVATE_WINDOW) {
     // Caller requested a private window
     isPrivateBrowsingWindow = true;
   }
-  if (aChromeMask & nsIWebBrowserChrome::CHROME_REMOTE_WINDOW) {
-    isUsingRemoteTabs = true;
-  }
-
   nsCOMPtr<mozIDOMWindowProxy> domWin = do_GetInterface(aParent);
   nsCOMPtr<nsIWebNavigation> webNav = do_GetInterface(domWin);
   nsCOMPtr<nsILoadContext> parentContext = do_QueryInterface(webNav);
@@ -709,17 +728,48 @@ nsresult nsAppShellService::JustCreateTopWindow(
     isPrivateBrowsingWindow = parentContext->UsePrivateBrowsing();
   }
 
-  if (parentContext) {
-    isUsingRemoteTabs = parentContext->UseRemoteTabs();
-  }
+  if (nsDocShell* docShell = window->GetDocShell()) {
+    MOZ_ASSERT(docShell->GetBrowsingContext()->IsChrome());
 
-  nsCOMPtr<mozIDOMWindowProxy> newDomWin =
-      do_GetInterface(NS_ISUPPORTS_CAST(nsIBaseWindow*, window));
-  nsCOMPtr<nsIWebNavigation> newWebNav = do_GetInterface(newDomWin);
-  nsCOMPtr<nsILoadContext> thisContext = do_GetInterface(newWebNav);
-  if (thisContext) {
-    thisContext->SetPrivateBrowsing(isPrivateBrowsingWindow);
-    thisContext->SetRemoteTabs(isUsingRemoteTabs);
+    docShell->SetPrivateBrowsing(isPrivateBrowsingWindow);
+    docShell->SetRemoteTabs(aChromeMask &
+                            nsIWebBrowserChrome::CHROME_REMOTE_WINDOW);
+    docShell->SetRemoteSubframes(aChromeMask &
+                                 nsIWebBrowserChrome::CHROME_FISSION_WINDOW);
+
+    // Eagerly create an about:blank content viewer with the right principal
+    // here, rather than letting it happening in the upcoming call to
+    // SetInitialPrincipalToSubject. This avoids creating the about:blank
+    // document and then blowing it away with a second one, which can cause
+    // problems for the top-level chrome window case. See bug 789773. Note that
+    // we don't accept expanded principals here, similar to
+    // SetInitialPrincipalToSubject.
+    if (nsContentUtils::IsInitialized()) {  // Sometimes this happens really
+                                            // early. See bug 793370.
+      nsCOMPtr<nsIPrincipal> principal =
+          nsContentUtils::SubjectPrincipalOrSystemIfNativeCaller();
+      if (nsContentUtils::IsExpandedPrincipal(principal)) {
+        principal = nullptr;
+      }
+      // Use the subject (or system) principal as the storage principal too
+      // until the new window finishes navigating and gets a real storage
+      // principal.
+      rv = docShell->CreateAboutBlankContentViewer(principal, principal,
+                                                   /* aCsp = */ nullptr);
+      NS_ENSURE_SUCCESS(rv, rv);
+      RefPtr<Document> doc = docShell->GetDocument();
+      NS_ENSURE_TRUE(!!doc, NS_ERROR_FAILURE);
+      doc->SetIsInitialDocument(true);
+    }
+
+    // Begin loading the URL provided.
+    if (aUrl) {
+      RefPtr<nsDocShellLoadState> loadState = new nsDocShellLoadState(aUrl);
+      loadState->SetTriggeringPrincipal(nsContentUtils::GetSystemPrincipal());
+      loadState->SetFirstParty(true);
+      rv = docShell->LoadURI(loadState, /* aSetNavigating */ true);
+      NS_ENSURE_SUCCESS(rv, rv);
+    }
   }
 
   window.forget(aResult);
@@ -731,9 +781,11 @@ nsresult nsAppShellService::JustCreateTopWindow(
 }
 
 NS_IMETHODIMP
-nsAppShellService::GetHiddenWindow(nsIXULWindow** aWindow) {
+nsAppShellService::GetHiddenWindow(nsIAppWindow** aWindow) {
   NS_ENSURE_ARG_POINTER(aWindow);
 
+  EnsureHiddenWindow();
+
   *aWindow = mHiddenWindow;
   NS_IF_ADDREF(*aWindow);
   return *aWindow ? NS_OK : NS_ERROR_FAILURE;
@@ -741,6 +793,10 @@ nsAppShellService::GetHiddenWindow(nsIXULWindow** aWindow) {
 
 NS_IMETHODIMP
 nsAppShellService::GetHiddenDOMWindow(mozIDOMWindowProxy** aWindow) {
+  NS_ENSURE_ARG_POINTER(aWindow);
+
+  EnsureHiddenWindow();
+
   nsresult rv;
   nsCOMPtr<nsIDocShell> docShell;
   NS_ENSURE_TRUE(mHiddenWindow, NS_ERROR_FAILURE);
@@ -755,38 +811,10 @@ nsAppShellService::GetHiddenDOMWindow(mozIDOMWindowProxy** aWindow) {
 }
 
 NS_IMETHODIMP
-nsAppShellService::GetHiddenPrivateWindow(nsIXULWindow** aWindow) {
-  NS_ENSURE_ARG_POINTER(aWindow);
+nsAppShellService::GetHasHiddenWindow(bool* aHasHiddenWindow) {
+  NS_ENSURE_ARG_POINTER(aHasHiddenWindow);
 
-  EnsurePrivateHiddenWindow();
-
-  *aWindow = mHiddenPrivateWindow;
-  NS_IF_ADDREF(*aWindow);
-  return *aWindow ? NS_OK : NS_ERROR_FAILURE;
-}
-
-NS_IMETHODIMP
-nsAppShellService::GetHiddenPrivateDOMWindow(mozIDOMWindowProxy** aWindow) {
-  EnsurePrivateHiddenWindow();
-
-  nsresult rv;
-  nsCOMPtr<nsIDocShell> docShell;
-  NS_ENSURE_TRUE(mHiddenPrivateWindow, NS_ERROR_FAILURE);
-
-  rv = mHiddenPrivateWindow->GetDocShell(getter_AddRefs(docShell));
-  NS_ENSURE_SUCCESS(rv, rv);
-  NS_ENSURE_TRUE(docShell, NS_ERROR_FAILURE);
-
-  nsCOMPtr<nsPIDOMWindowOuter> hiddenPrivateDOMWindow(docShell->GetWindow());
-  hiddenPrivateDOMWindow.forget(aWindow);
-  return *aWindow ? NS_OK : NS_ERROR_FAILURE;
-}
-
-NS_IMETHODIMP
-nsAppShellService::GetHasHiddenPrivateWindow(bool* aHasPrivateWindow) {
-  NS_ENSURE_ARG_POINTER(aHasPrivateWindow);
-
-  *aHasPrivateWindow = !!mHiddenPrivateWindow;
+  *aHasHiddenWindow = !!mHiddenWindow;
   return NS_OK;
 }
 
@@ -800,7 +828,7 @@ nsAppShellService::GetApplicationProvidedHiddenWindow(bool* aAPHW) {
  * Register a new top level window (created elsewhere)
  */
 NS_IMETHODIMP
-nsAppShellService::RegisterTopLevelWindow(nsIXULWindow* aWindow) {
+nsAppShellService::RegisterTopLevelWindow(nsIAppWindow* aWindow) {
   NS_ENSURE_ARG_POINTER(aWindow);
 
   nsCOMPtr<nsIDocShell> docShell;
@@ -809,7 +837,7 @@ nsAppShellService::RegisterTopLevelWindow(nsIXULWindow* aWindow) {
 
   nsCOMPtr<nsPIDOMWindowOuter> domWindow(docShell->GetWindow());
   NS_ENSURE_TRUE(domWindow, NS_ERROR_FAILURE);
-  domWindow->SetInitialPrincipalToSubject();
+  domWindow->SetInitialPrincipalToSubject(nullptr, Nothing());
 
   // tell the window mediator about the new window
   nsCOMPtr<nsIWindowMediator> mediator(
@@ -832,15 +860,15 @@ nsAppShellService::RegisterTopLevelWindow(nsIXULWindow* aWindow) {
 
   if (obssvc) {
     obssvc->NotifyObservers(aWindow, "xul-window-registered", nullptr);
-    nsXULWindow* xulWindow = static_cast<nsXULWindow*>(aWindow);
-    xulWindow->WasRegistered();
+    AppWindow* appWindow = static_cast<AppWindow*>(aWindow);
+    appWindow->WasRegistered();
   }
 
   return NS_OK;
 }
 
 NS_IMETHODIMP
-nsAppShellService::UnregisterTopLevelWindow(nsIXULWindow* aWindow) {
+nsAppShellService::UnregisterTopLevelWindow(nsIAppWindow* aWindow) {
   if (mXPCOMShuttingDown) {
     /* return an error code in order to:
        - avoid doing anything with other member variables while we are in
@@ -859,10 +887,6 @@ nsAppShellService::UnregisterTopLevelWindow(nsIXULWindow* aWindow) {
     // CreateHiddenWindow() does not register the window, so we're done.
     return NS_OK;
   }
-  if (aWindow == mHiddenPrivateWindow) {
-    // CreateHiddenWindow() does not register the window, so we're done.
-    return NS_OK;
-  }
 
   // tell the window mediator
   nsCOMPtr<nsIWindowMediator> mediator(
@@ -897,9 +921,6 @@ nsAppShellService::Observe(nsISupports* aSubject, const char* aTopic,
     if (mHiddenWindow) {
       mHiddenWindow->Destroy();
     }
-    if (mHiddenPrivateWindow) {
-      mHiddenPrivateWindow->Destroy();
-    }
   } else {
     NS_ERROR("Unexpected observer topic!");
   }
